"""A parser for T602 files"""

# This file is based on HTMLParser.py

import re

bold_toggle = chr(0x02)
italics_toggle = chr(0x04)
sup_toggle = chr(0x14)
sub_toggle = chr(0x16)
wide_toggle = chr(0x0f)
tall_toggle = chr(0x10)
big_toggle = chr(0x1d)
underline_toggle = chr(0x13)
hard_space = chr(0xfe)

soft_space = ' '

eof = chr(0x1a)

font_elite = chr(0x01)
font_condens = chr(0x03)
font_uzv1 = chr(0x11)
font_uzv2 = chr(0x12)
font_uzv3 = chr(0x15)
font_uzv4 = chr(0x17)
font_uzv5 = chr(0x18)
font_uzv6 = chr(0x19)

soft_eol = chr(0x8d)+chr(0x0a)
hard_eol = chr(0x0d)+chr(0x0a)

linespread_tag = '@LH'
newpage_tag = '@PA'
linelength_tag = '@PL'
left_margin = '@LM'
right_margin = '@RM'
top_margin = '@MT'
bottom_margin = '@MB'
header_tag = '@HE'
footer_tag = '@FO'
tabs = '@TB'

po_tag = '@PO'
pn_tag = '@PN'
op_tag = '@OP'
cp_tag = '@CP'

codepage = '@CT'
codepages = {0:'kamenicky', 1:'cp852', 2:'koi8_cs602'}

picture_tag = '.PI'

tag_regexp = re.compile("^[@\.][A-Z][A-Z](\ |$)")

class T602ParseError(Exception):
    """Exception raised for all parse errors."""

    def __init__(self, msg, position=(None, None)):
        assert msg
        self.msg = msg
        self.lineno = position[0]
        self.offset = position[1]

    def __str__(self):
        result = self.msg
        if self.lineno is not None:
            result = result + ", at line %d" % self.lineno
        if self.offset is not None:
            result = result + ", column %d" % (self.offset + 1)
        return result


class T602Parser:
    """Find markup and call handler functions.

    Usage:
        p = T602Parser()
        p.feed(data)
        ...
        p.close()
    """

    def __init__(self):
        """Initialize and reset this instance."""
        self.reset()

    def reset(self):
        """Reset this instance.  Loses all unprocessed data."""
        self.rawdata = ''
        
        self.sup = 0
        self.sub = 0
        
        self.bold = 0
        self.italics = 0
        self.underline = 0
        self.wide = 0
        self.tall = 0
        self.big = 0

        self.codepage = None
        self.leftmargin = None
        self.rightmargin = None
        self.topmargin = None
        self.bottommargin = None

        self.linespread = None
        self.linelength = None       

        self.header = None
        self.footer = None

        self.tabs = None

        self.encoding = None

        self.eof_encountered = 0

    def feed(self, data):
        """Feed data to the parser.

        Call this as often as you want, with as little or as much text
        as you want (may include '\n').
        """
        if self.eof_encountered:
            return
        self.rawdata = self.rawdata + data
        self.goahead(0)

    def close(self):
        """Handle any buffered data."""
        self.goahead(1)

    def error(self, message):
        raise T602ParseError(message)


    def process_tag(self, tag, params):
        if tag==linespread_tag:
            self.linespread = params
        elif tag==newpage_tag:
            self.handle_newpage()
        elif tag==linelength_tag:
            self.linelength = params
        elif tag==left_margin:
            self.leftmargin = params
        elif tag==right_margin:
            self.rightmargin = params
        elif tag==top_margin:
            self.topmargin = params
        elif tag==bottom_margin:
            self.bottommargin = params
        elif tag==header_tag:
            self.header = params
        elif tag==footer_tag:
            self.footer = params
        elif tag==tabs:
            self.tabs = params
        elif tag==codepage:
            self.encoding = codepages[int(params)] 
        elif tag==po_tag:
            pass
        elif tag==pn_tag:
            pass
        elif tag==op_tag:
            pass
        elif tag==cp_tag:
            pass
        else:
            raise T602ParseError("Unknown tag "+tag)

    def process_bold_toggle(self):
        self.bold = not self.bold
        self.handle_bold_toggle()
        if self.bold:
            self.handle_bold_on()
        else:
            self.handle_bold_off()

    def process_italics_toggle(self):
        self.italics = not self.italics
        self.handle_italics_toggle()
        if self.italics:
            self.handle_italics_on()
        else:
            self.handle_italics_off()

    def process_sup_toggle(self):
        self.sup = not self.sup
        self.handle_sup_toggle()
        if self.sup:
            self.handle_sup_on()
        else:
            self.handle_sup_off()

    def process_sub_toggle(self):
        self.sub = not self.sub
        self.handle_sub_toggle()
        if self.sub:
            self.handle_sub_on()
        else:
            self.handle_sub_off()

    def process_wide_toggle(self):
        self.wide = not self.wide
        self.handle_wide_toggle()
        if self.wide:
            self.handle_wide_on()
        else:
            self.handle_wide_off()

    def process_tall_toggle(self):
        self.tall = not self.tall
        self.handle_tall_toggle()
        if self.tall:
            self.handle_tall_on()
        else:
            self.handle_tall_off()

    def process_big_toggle(self):
        self.big = not self.big
        self.handle_big_toggle()
        if self.big:
            self.handle_big_on()
        else:
            self.handle_big_off()

    def process_underline_toggle(self):
        self.underline = not self.underline
        self.handle_underline_toggle()
        if self.underline:
            self.handle_underline_on()
        else:
            self.handle_underline_off()

    

    def process_control(self, c):
        dc = {
        font_elite      : lambda x: self.handle_font_change('elite'),
        bold_toggle     : self.process_bold_toggle,
        font_condens    : lambda x: self.handle_font_change('condens'),
        italics_toggle  : self.process_italics_toggle,
        font_uzv1       : lambda x: self.handle_font_change('uzv1'),
        font_uzv2       : lambda x: self.handle_font_change('uzv2'),
        font_uzv3       : lambda x: self.handle_font_change('uzv3'),
        font_uzv4       : lambda x: self.handle_font_change('uzv4'),
        font_uzv5       : lambda x: self.handle_font_change('uzv5'),
        font_uzv6       : lambda x: self.handle_font_change('uzv9'),
        
        sup_toggle      : self.process_sup_toggle,
        sub_toggle      : self.process_sub_toggle,
        wide_toggle     : self.process_wide_toggle,
        tall_toggle     : self.process_tall_toggle,
        big_toggle      : self.process_big_toggle,
        underline_toggle: self.process_underline_toggle,
        } 
        if dc.has_key(c):
            dc[c]()
        else:
            self.handle_invalid_cc(c)


    def process_soft_spaces(self, number):
        self.handle_soft_spaces(number)

    # internal - process outgoing data
    def process_out_data(self, data):
        self.handle_data(unicode(data, self.encoding))

    # Internal - process one line 
    # returns True if line was real line with data
    # returns False if line just contained formatting tag
    def process_line(self, line):
        rx = re.compile("^@[A-Z][A-Z](\ |$)")
        if re.search(rx, line):
            self.process_tag(line[:3], line[4:])
            return 0
        outdata = ''
        i = 0
        linelen = len(line)
        while i<linelen:
            c = line[i]
            if c<' ':
                if outdata:
                    self.process_out_data(outdata)
                    outdata = ''
                if c==eof:
                    self.eof_encountered = 1
                    self.handle_eof()                    
                    return 1
                i+=1
                self.process_control(c)
                continue
            else:
                if line[i:i+2]==soft_eol:
                    i+=2
                    if outdata:
                        self.process_out_data(outdata)
                        outdata = ''
                    self.handle_soft_eol()
                    continue
                if c==soft_space:
                    if outdata:
                        self.process_out_data(outdata)
                        outdata = ''
                    nsoftspaces = 1
                    i+=1
                    while line[i]==soft_space:
                        i+=1
                        nsoftspaces += 1
                    self.process_soft_spaces(nsoftspaces)
                    continue
                if c==hard_space:
                    i+=1
                    if outdata:
                        self.process_out_data(outdata)
                        outdata = ''
                    self.handle_hard_space()
                    continue
                i+=1
                outdata += c
        if outdata:
            self.process_out_data(outdata)
        return 1
                
                        

    # Internal -- handle data as far as reasonable.  May leave state
    # and data to be processed by a subsequent call.  If 'end' is
    # true, force handling all data as if followed by EOF marker.
    def goahead(self, end):
        rawdata = self.rawdata
        lines = rawdata.split(hard_eol)
        for line in lines[:-1]:
            was_data = self.process_line(line)
            if was_data:
                self.handle_hard_eol()
        self.rawdata = lines[-1]
        if end:
            self.process_line(self.rawdata)
            


    # Overridable -- handle soft end of line
    def handle_soft_eol(self):
        self.handle_whitespace()

    # Overridable -- handle hard end of line
    def handle_hard_eol(self):
        self.handle_data(u'\n')


    # Overridable -- handle soft spaces
    def handle_soft_spaces(self, number):
        self.handle_whitespace()

    # Overridable -- handle hard space
    def handle_hard_space(self):
        self.handle_data(u'\u00a0') # U+00A0 NO-BREAK SPACE

    # Overridable -- handle new page
    def handle_newpage(self):
        pass

    # Overridable -- handle data
    def handle_data(self, data):
        pass

    # Overridable -- handle whitespace
    def handle_whitespace(self):
        self.handle_data(u' ')

    # Overridable -- handle end of file
    def handle_eof(self):
        pass

    # Overridable -- handle font change
    def handle_font_change(self, font):
        pass

    # Overridable -- handle superscript toggle
    def handle_sup_toggle(self):
        pass

    # Overridable -- handle superscript turned on
    def handle_sup_on(self):
        pass

    # Overridable -- handle superscript turned off
    def handle_sup_off(self):
        pass

 
    # Overridable -- handle subscript toggle
    def handle_sub_toggle(self):
        pass

    # Overridable -- handle subcript turned on
    def handle_sub_on(self):
        pass

    # Overridable -- handle subscript turned off
    def handle_sub_off(self):
        pass


    # Overridable -- handle bold toggle
    def handle_bold_toggle(self):
        pass

    # Overridable -- handle bold turned on
    def handle_bold_on(self):
        pass

    # Overridable -- handle bold turned off
    def handle_bold_off(self):
        pass


    # Overridable -- handle italics toggle
    def handle_italics_toggle(self):
        pass

    # Overridable -- handle italics turned on
    def handle_italics_on(self):
        pass

    # Overridable -- handle italics turned off
    def handle_italics_off(self):
        pass


    # Overridable -- handle wide toggle
    def handle_wide_toggle(self):
        pass

    # Overridable -- handle wide turned on
    def handle_wide_on(self):
        pass

    # Overridable -- handle wide turned off
    def handle_wide_off(self):
        pass

 
    # Overridable -- handle tall toggle
    def handle_tall_toggle(self):
        pass

    # Overridable -- handle tall turned on
    def handle_tall_on(self):
        pass

    # Overridable -- handle tall turned off
    def handle_tall_off(self):
        pass

 
    # Overridable -- handle big toggle
    def handle_big_toggle(self):
        pass

    # Overridable -- handle big turned on
    def handle_big_on(self):
        pass

    # Overridable -- handle big turned off
    def handle_big_off(self):
        pass


    # Overridable -- handle underline toggle
    def handle_underline_toggle(self):
        pass

    # Overridable -- handle underline turned on
    def handle_underline_on(self):
        pass

    # Overridable -- handle underline turned off
    def handle_underline_off(self):
        pass


    # Overridable -- handle invalid control character
    def handle_invalid_cc(self, c):
        self.error("Invalid control character 0x%02x" % ord(c))
        

                
if __name__=='__main__':
    import sys
    class My602Parser(T602Parser):
    
        def handle_data(self, data):
            sys.stdout.write(data.encode('utf-8'))

    p = My602Parser()
    f = open(sys.argv[1])
    for i in f.readlines():
        p.feed(i)
    p.close()
    sys.stdout.write('\n')
